/*
 * babyloader
 *
 * Copyright (C) 2005 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <config.h>
#include <turbo/turbo.h>

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "image.h"
#include "pins.h"
#include "isp.h"
#include "delay.h"

/* *INDENT-OFF* */

lc_char PROGMEM lc_Isp_Info[]={
	LC_EN("Babyloader")
	LC_END
};

lc_char PROGMEM lc_Info[]={
	LC_EN("Info")
	LC_END
};

lc_char PROGMEM lc_Update[]={
	LC_EN("Update")
	LC_END
};

lc_char PROGMEM lc_About[]={
	LC_EN("About")
	LC_END
};

/* *INDENT-ON* */

u8 PROGMEM t_y[] = "Y\n";
u8 PROGMEM t_n[] = "N\n";

u8 PROGMEM t_initializing[] = "Initializing...\n";
u8 PROGMEM t_enabling[] = "Enabling...\n";
u8 PROGMEM t_reading[] = "Reading...\n";
u8 PROGMEM t_writingfuses[] = "Writing fuses...\n";
u8 PROGMEM t_erasing[] = "Erasing...\n";
u8 PROGMEM t_updating[] = "Updating...\n";
u8 PROGMEM t_verifing[] = "Verifing...\n";
u8 PROGMEM t_verified[] = "Verified...\n";
u8 PROGMEM t_locking[] = "Locking...\n";
u8 PROGMEM t_updated[] = "Updated\n";

u8 PROGMEM t_part[] = "Part=";
u8 PROGMEM t_lock[] = "Lock=";
u8 PROGMEM t_fulow[] = "FuseLow=";
u8 PROGMEM t_fuhigh[] = "FuseHigh=";
u8 PROGMEM t_fuext[] = "FuseExt=";
u8 PROGMEM t_calib[] = "Calib=";
u8 PROGMEM t_flash[] = "FLASH=";
u8 PROGMEM t_image[] = "IMAGE=";

//-------------------------------------------------
// Initialization of variables
//-------------------------------------------------
void init_variables (void)
{

}

//-------------------------------------------------
// MCU's ports initialization
//-------------------------------------------------
void init_mcuPorts (void)
{
  //ISP_RESETn
  //change direction to output
  //set pin to log.1
  //
  sbi (ISP_RESETn_DDR, ISP_RESETn_DDR_DD);
  sbi (ISP_RESETn_PORT, ISP_RESETn_PORT_PIN);

  //ISP_SCK
  //let it set to input
  //
  //sbi(DDRB,DD1);

  //ISP_MOSI
  //let it set to input
  //

  //ISP_MISO
  //change direction to output
  //set pin to log.1
  //
  sbi (ISP_MISO_DDR, ISP_MISO_DDR_DD);
  sbi (ISP_MISO_PORT, ISP_MISO_PORT_PIN);

  //ISP_SS 
  //change direction to output
  //set pin to log.1
  sbi (ISP_SS_DDR, ISP_SS_DDR_DD);
  sbi (ISP_SS_PORT, ISP_SS_PORT_PIN);
}

//-------------------------------------------------
// SPI iface initialization
//-------------------------------------------------
void init_Spi (void)
{
//  // enable SPI, set SPI MASTER mode
//  // CLKosc/128 SPI clock (with SPI2X /64)
//  outb (SPCR, 0x53);
//  // Disable double SPI clock
//  cbi (SPSR, SPI2X);
  // enable SPI, set SPI MASTER mode
  // CLKosc/4 SPI clock (with SPI2X /2)
  outb (SPCR, 0x50);
  // Enable double SPI clock
  sbi (SPSR, SPI2X);
}

//-------------------------------------------------
//-------------------------------------------------
u8 init_Isp ()
{
  // disable SPI interrupt
  cbi (SPCR, SPIE);
  // disable SPI
  cbi (SPCR, SPE);
  // clear SPI interrupt flag
  inb (SPSR);
  inb (SPDR);

//  // enable SPI, set SPI MASTER mode
//  // CLKosc/128 SPI clock (with SPI2X /64)
//  outb (SPCR, 0x53);
//  // Disable double SPI clock
//  cbi (SPSR, SPI2X);
  // enable SPI, set SPI MASTER mode
  // CLKosc/16 SPI clock (with SPI2X /8)
  outb (SPCR, 0x51);
  // Disable double SPI clock
  cbi (SPSR, SPI2X);

  //ISP_SS change direction to output 
  sbi (ISP_SS_DDR, ISP_SS_DDR_DD);

  // RESETn pin stays output and set it to log.0
  cbi (ISP_RESETn_PORT, ISP_RESETn_PORT_PIN);
  sbi (ISP_RESETn_DDR, ISP_RESETn_DDR_DD);

  // wait for a moment
  delayMs (20);

  // SCK pin will be output and set it to log.0
  cbi (ISP_SCK_PORT, ISP_SCK_PORT_PIN);
  sbi (ISP_SCK_DDR, ISP_SCK_DDR_DD);

//      // PDI pin stays output and set it to log.1
//      sbi(ISP_PDI_PORT,ISP_PDI_PORT_PIN);
//        sbi(ISP_PDI_DDR,ISP_PDI_DDR_DD);

  // MOSI pin is output and set it to log.1
  sbi (ISP_MOSI_PORT, ISP_MOSI_PORT_PIN);
  sbi (ISP_MOSI_DDR, ISP_MOSI_DDR_DD);

//      // PDO stays input
//        cbi(ISP_PDO_DDR,ISP_PDO_DDR_DD);

  // MISO is input
  cbi (ISP_MISO_DDR, ISP_MISO_DDR_DD);

  isp_out_reset (0);
  isp_out_sck (0);
  delayMs (20);
  isp_pulse_reset ();
  delayMs (20);

  return 1;
}

//-------------------------------------------------
// Top initialization function
//-------------------------------------------------
void init_top (void)
{
//TODO
//    // disable interrupts globaly
//    cli(); 

//    dbsp ("ISP init_top\n");
  //initialize variables
  init_variables ();
  //initialize all MCU ports
  init_mcuPorts ();
//  //initialize SPI iface
//  init_Spi ();
  //initialize ISP
  init_Isp ();

//TODO
//    // enable interrupts globaly
//    sei();
}

u8 display_text_x (u8 * text)
{
  u8 ret = display_text (text, NULL);

  if (ret == APP_END)
    return APP_END;
  return APP_BACK;
}

u8 babyloader_info (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 *t = buf_B ();
    u8 *r;
    u8 i;
    u8 tmpVal;
    u8 Lock;
    u8 FuLow;
    u8 FuHigh;
    u8 FuExt;
    u8 Calib;
    u8 flash[16];
    u8 Part[4];
    u32 page_number;
    u32 page_offset;
    u32 verifyErrors;

    init_top ();

    delayMs (20);
    isp_enable_avr ();

    r = t;

    //read Lock,Fuse,Calib bytes
    r = sprints (r, t_part);
    for (i = 0; i < 3; i++)
      Part[i] = isp_get_part_info (i);
    for (i = 0; i < 3; i++)
    {
      r = sprintch (r, Part[i]);
      r = sprintc (r, ' ');
    }
    r = sprintc (r, '\n');

    r = sprints (r, t_lock);
    tmpVal = isp_read_lock_bits ();
    Lock = tmpVal;
    r = sprintch (r, tmpVal);
    r = sprintc (r, '\n');

    r = sprints (r, t_fulow);
    tmpVal = isp_read_fuse_low_bits ();
    FuLow = tmpVal;
    r = sprintch (r, tmpVal);
    r = sprintc (r, '\n');

    r = sprints (r, t_fuhigh);
    tmpVal = isp_read_fuse_high_bits ();
    FuHigh = tmpVal;
    r = sprintch (r, tmpVal);
    r = sprintc (r, '\n');

    r = sprints (r, t_fuext);
    tmpVal = isp_read_fuse_ext_bits ();
    FuExt = tmpVal;
    r = sprintch (r, tmpVal);
    r = sprintc (r, '\n');

    r = sprints (r, t_calib);
    tmpVal = isp_read_cal_byte (0);
    Calib = tmpVal;
    r = sprintch (r, tmpVal);
    r = sprintc (r, '\n');

    r = sprints (r, t_flash);
    isp_read_chunk_flash (0, flash, 8);
    for (i = 0; i < 8; i++)
    {
      r = sprintch (r, flash[i]);
      r = sprintc (r, ' ');
    }
    r = sprintc (r, '\n');

    r = sprints (r, t_image);
    for (i = 0; i < 8; i++)
    {
      tmpVal = rb (baby_image + i);
      r = sprintch (r, tmpVal);
      r = sprintc (r, ' ');
    }
    r = sprintc (r, '\n');

    r = sprintc (r, '\0');
    return display_text_x (t);
  }
  return APP_OK;
}

u16 update ()			// rets #errors
{
  u8 i;
  u8 tmpVal;
  u8 Lock;
  u8 FuLow;
  u8 FuHigh;
  u8 FuExt;
  u8 Calib;
  u8 flash[16];
  u8 Part[4];
  u16 page_number;
  u16 page_offset;
  u16 verifyErrors;
  u16 baby_size = sizeof (baby_image);

  init_top ();

  delayMs (20);
  isp_enable_avr ();

  //erase chip
  delayMs (20);
  isp_chip_erase ();

  //enable programming
  delayMs (20);
  isp_enable_avr ();

  delayMs (20);

  isp_write_fuse_low_bits (0x67);
  isp_write_fuse_high_bits (0xDF);
  isp_write_fuse_ext_bits (0xFF);

  //start programming
  delayMs (20);
  for (page_number = 0; page_number < 128; page_number++)
  {
    for (page_offset = 0; page_offset < 128; page_offset++)
    {
      if ((page_number << 7) + page_offset < baby_size)
	tmpVal = rb (baby_image + (page_number << 7) + page_offset);
      else
	tmpVal = 0xff;

      isp_write_byte_flash (page_offset, tmpVal);
    }
    isp_write_program_memory_page ((page_number << 7));
    if ((page_number << 7) > baby_size)
      break;
  }

  //start verifing
  delayMs (20);
  verifyErrors = 0;
  for (page_number = 0; page_number < 128; page_number++)
  {
    for (page_offset = 0; page_offset < 128; page_offset++)
    {
      if ((page_number << 7) + page_offset < baby_size)
	tmpVal = rb (baby_image + (page_number << 7) + page_offset);
      else
	tmpVal = 0xff;
      if (isp_read_byte_flash
	  ((unsigned long) ((page_number << 7) + page_offset)) != tmpVal)
      {
	verifyErrors++;
      }
    }
    if ((page_number << 7) > baby_size)
      break;
  }

  isp_write_lock_bits (0xC0);
  delayMs (20);

  return verifyErrors;
}

u8 babyloader_update (SCtx * ctx, u8 action)
{
  if (action == APP_ENTER)
  {
    u8 *t = buf_B ();
    u8 *r;
    u16 verifyErrors;

    dbsp ("UPDATE_START\n");

    verifyErrors = update ();

    r = t;
    r = sprints (r, t_verified);
    r = sprintc (r, '\n');
    r = sprints (r, "Errors: ");
    r = sprinti (r, verifyErrors);
    r = sprintc (r, '\n');

    r = sprintc (r, '\0');

//#if 0
    //Start 88
    isp_out_reset (1);
    delayMs (200);
    isp_out_reset (0);
    delayMs (200);
    isp_out_reset (1);
    delayMs (200);
//#endif

    return display_text_x (t);
  }
  return APP_OK;
}

u8 babyloader_about(SCtx * ctx, u8 action)
{
	if(action==APP_ENTER)
	{
		u8 *res;
		u8 * t=buf_B();
		u8 *r=t;

		//VERSION
		r=sprinti(r,TURBO_VERSION_A);
		r=sprintc(r,'.');
		r=sprinti(r,TURBO_VERSION_B);
		r=sprintc(r,'.');
		r=sprinti(r,TURBO_VERSION_C);
/*
		r=sprintc(r,'.');
		r=sprinti(r,TURBO_VERSION_D);
*/
		r=sprintc(r,'\n');
		r=sprintc(r,'\0');
		return display_text_x(t);
	}
return APP_OK;
}

SNodeP n_1 = { lc_Isp_Info, NULL };
SNodeP n_2_1 = { lc_Info, babyloader_info };
SNodeP n_2_2 = { lc_Update, babyloader_update };
SNodeP n_2_3 = { lc_About, babyloader_about };

/* *INDENT-OFF* */
SEdgeP isp_edges_p[] = {
  {&n_1, &n_2_1},		// ISP/Info
  {&n_1, &n_2_2},		// ISP/Update
  {&n_1, &n_2_3},		// ISP/About
  NULL
};
/* *INDENT-ON* */

void action_menu (void *data)
{
  SCtx *c = spider_init ();

  if (c == NULL)
    return;

  c->n = &n_1;
  c->eP = &isp_edges_p;

  spider (c);
}

// !!!
//TODO STORE AND RESTORE SPI STATE due to debug or mmc!!!!
// !!!
void turbo_handler (u8 action, void *data)
{
  switch (action)
  {
    case ACTION_APP_REGISTER:
      update ();
      break;
    case ACTION_INSERT_MENU:
      insert_menu (locale (lc_Isp_Info));
      break;
    case ACTION_MENU_SELECTION:
      stk_thread (action_menu, data);
      break;
    default:
      break;
  }
}
