#include <config.h>
#include <turbo/turbo.h>

#include "delay.h"
#include "pins.h"
#include "isp.h"

#include <avr/io.h>

void isp_sck_delay ()
{
//TRTRTR    delayUs(3);
  delayUs (10);
}

void isp_out_reset (u8 resetValue)
{
  u8 tmpResetValue;

  tmpResetValue = inb (ISP_RESETn_PORT);
  if (resetValue == 0)
  {
    tmpResetValue = tmpResetValue & ~(1 << ISP_RESETn_PORT_PIN);
  }
  else
  {
    tmpResetValue = tmpResetValue | (1 << ISP_RESETn_PORT_PIN);
  }
  outb (ISP_RESETn_PORT, tmpResetValue);
}

void isp_out_sck (u8 sckValue)
{
  u8 tmpSckValue;

  tmpSckValue = inb (ISP_SCK_PORT);
  if (sckValue == 0)
  {
    tmpSckValue = tmpSckValue & ~(1 << ISP_SCK_PORT_PIN);
  }
  else
  {
    tmpSckValue = tmpSckValue | (1 << ISP_SCK_PORT_PIN);
  }
  outb (ISP_SCK_PORT, tmpSckValue);
}

u8 isp_send_recv (u8 b)
{
  // clear flag
  inb (SPSR);
  inb (SPDR);

  //disable debugging
  cbi (SPI_SSn_PORT, SPI_SSn_PORT_PIN);

  cbi (ISP_SS_PORT, ISP_SS_PORT_PIN);
//  delayUs (10);
  delayUs (1);

  // send char to SPI
  outb (SPDR, b);
  // wait until char was transmitted
  loop_until_bit_is_set (SPSR, SPIF);

  //enable debugging
  sbi (SPI_SSn_PORT, SPI_SSn_PORT_PIN);

  sbi (ISP_SS_PORT, ISP_SS_PORT_PIN);
//  delayUs (10);
  delayUs (1);

  // return received value
  return inb (SPDR);
}

u16 isp_send (u8 * queue, u16 queueSize)
{
  u8 *p = queue, ch;
  u16 i = queueSize;

  while (i--)
  {
    ch = isp_send_recv (*p);
    *p++ = ch;
  }
  return queueSize;
}

void isp_pulse_sck (void)
{
  isp_sck_delay ();		//TRTRTR
//TRTRTR    __asm__ __volatile__ ("nop");
//TRTRTR    __asm__ __volatile__ ("nop");
  isp_out_sck (1);
  isp_sck_delay ();		//TRTRTR
//TRTRTR    __asm__ __volatile__ ("nop");
//TRTRTR    __asm__ __volatile__ ("nop");
  isp_out_sck (0);
}

void isp_pulse_reset (void)
{
  isp_out_reset (1);
  isp_out_reset (0);
}

u8 isp_enable_avr (void)
{
  u8 prg[4];
  b8 try_number = 32;

  do
  {
    prg[0] = 0xAC;
    prg[1] = 0x53;
    prg[2] = prg[3] = 0;
    isp_send (prg, 4);
    if (prg[2] == 0x53)
    {
      break;
    }
    isp_pulse_sck ();		//TRTRTR
//TRTRTR    __asm__ __volatile__ ("nop");
//TRTRTR    __asm__ __volatile__ ("nop");
//TRTRTR    __asm__ __volatile__ ("nop");
  }
  while (try_number--);

  if (try_number > 0)
  {
    //OK
    return 1;
  }
  else
  {
    //ERROR
    return 0;
  }
}

u8 isp_get_part_info (u8 addr)
{
  u8 info[4] = { 0x30, 0, addr, 0 };

  isp_send (info, 4);
  return info[3];
}

u8 isp_read_byte_flash (u32 addr)
{
  u8 hl = (addr & 1) ? 0x28 : 0x20;
  u8 flash[4] =
    { hl, (u8) ((addr >> 9) & 0xFF), (u8) ((addr >> 1) & 0xFF), 0 };

  isp_send (flash, 4);
  return flash[3];
}

void isp_read_chunk_flash (u32 addr, u8 * chunk, u8 chunkSize)
{
  u8 i;

  for (i = 0; i < chunkSize; i++)
  {
    chunk[i] = isp_read_byte_flash (addr);
    addr++;
  }
}

u8 isp_read_byte_eeprom (u16 addr)
{
  u8 eeprom[4] = { 0xA0, (u8) ((addr >> 8) & 0xFF), (u8) (addr & 0xFF), 0 };

  isp_send (eeprom, 4);
  return eeprom[3];
}

void isp_read_chunk_eeprom (u16 addr, u8 * chunk, u8 chunkSize)
{
  u8 i;

  for (i = 0; i < chunkSize; i++)
  {
    chunk[i] = isp_read_byte_eeprom (addr);
    addr++;
  }
}

u8 isp_read_lock_bits ()
{
  u8 lockfuse[4] = { 0x58, 0, 0, 0 };

  isp_send (lockfuse, 4);
  return lockfuse[3];
}

u8 isp_read_fuse_low_bits ()
{
  u8 fuselow[4] = { 0x50, 0, 0, 0 };

  isp_send (fuselow, 4);
  return fuselow[3];
}

u8 isp_read_fuse_high_bits ()
{
  u8 fusehigh[4] = { 0x58, 0x08, 0, 0 };

  isp_send (fusehigh, 4);
  return fusehigh[3];
}

u8 isp_read_fuse_ext_bits ()
{
  u8 fuseext[4] = { 0x50, 0x08, 0, 0 };

  isp_send (fuseext, 4);
  return fuseext[3];
}

u8 isp_read_cal_byte (u8 addr)
{
  u8 cal[4] = { 0x38, 0, addr, 0 };

  isp_send (cal, 4);
  return cal[3];
}

void isp_write_old_fuse_bits (u8 val)
{
  u8 oldfuse[4] = { 0xAC, (val & 0x1F) | 0xA0, 0, 0xD2 };
  isp_send (oldfuse, 4);
}

void isp_write_fuse_low_bits (u8 val)
{
  u8 fuselow[4] = { 0xAC, 0xA0, 0, val };
  isp_send (fuselow, 4);
}

void isp_write_fuse_high_bits (u8 val)
{
  u8 fusehigh[4] = { 0xAC, 0xA8, 0, val };
  isp_send (fusehigh, 4);
}

void isp_write_fuse_ext_bits (u8 val)
{
  u8 fuseext[4] = { 0xAC, 0xA4, 0, val };
  isp_send (fuseext, 4);
}

void isp_chip_erase ()
{
  u8 chip_erase[4] = { 0xAC, 0x80, 0x00, 0x00 };
  isp_send (chip_erase, 4);
  delayUs (9000);
  delayUs (9000);
  isp_pulse_reset ();
  isp_enable_avr ();
}

void isp_write_lock_bits (u8 bits)
{
  u8 lock[4] = { 0xAC, 0xF9 | ((bits << 1) & 0x06), 0xFF, bits };
  u8 rbits;

  isp_send (lock, 4);
  delayUs (9000);
  isp_pulse_reset ();
  isp_enable_avr ();
}

void isp_write_byte_flash (u16 addr, u8 val)
{
  u8 flash[4] =
    { 0x40 | ((addr & 0x01) << 3), 0, (u8) ((addr >> 1) & 0xFF), val };
  isp_send (flash, 4);
}

void isp_write_chunk_flash (u16 addr, u8 * chunk, u8 chunkSize)
{
  u8 j;

  for (j = 0; j < chunkSize; j++)
  {
    isp_write_byte_flash (addr, chunk[j]);
    addr++;
  }
}

// 0 ok
// 1 error
u8 isp_write_byte_eeprom (u16 addr, u8 val)
{
  u8 readback;
  u8 eeprom[4] = { 0xC0, (addr >> 8), (addr & 0xFF), val };
  u16 j;

  readback = isp_read_byte_eeprom (addr);
  if (readback == val)
    return 0;

  isp_send (eeprom, 4);
  for (j = 0; j < 100; j++)
  {
    delayUs (1000);
    readback = isp_read_byte_eeprom (addr);
    if (readback == val)
      j = 100;
  }

  if (readback != val)
    return 1;
  return 0;
}

// 1 OK
// 0 ERROR
u8 isp_write_chunk_eeprom (u16 addr, u8 * chunk, u8 chunkSize)
{
  u8 j;
  u8 err;

  for (j = 0; j < chunkSize; j++)
  {
    err = isp_write_byte_eeprom (addr, chunk[j]);
    if (err == 1)
      return 0;
    addr++;
  }
  return 1;
}

void isp_write_program_memory_page (u16 pageAddr)
{
  u8 prg_page[4] =
    { 0x4C, (u8) ((pageAddr >> 9) & 0xFF), (u8) ((pageAddr >> 1) & 0xFF), 0 };
//  u8 prg_page [4] = { 0x4C,(u8)((pageAddr >> 10) & 0x1F),(u8)(((pageAddr >> 6) & 0x0F)<<4),0 };
//  u8 prg_page [4] = { 0x4C,(u8)((pageAddr >> 9) & 0x1F),(u8)(((pageAddr >> 5) & 0x0E)<<5),0 };
//  u8 prg_page [4] = { 0x4C,(u8)((pageAddr >> 8) & 0x1F),(u8)(((pageAddr >> 4) & 0x0C)<<6),0 };
  isp_send (prg_page, 4);

  delayUs (5000);
//    for(j=0;j<20;j++) {
//      readback=ispReadByteFlash(tmpAddr);
//      if(readback == chunk[i]) j=20;
//    }
}
